
static char rcsid[] = "$Id: g_bird.c,v 1.3 1995/05/23 05:25:59 nakano Exp nakano $";

#include "g_private.h"
#include "g_bird.h"

static SCALE		saved_scale;
static LINE_PROP	saved_line;
static AREA_PROP	saved_area;
static G_POINT		saved_point;
static unsigned long	saved_bw_color_mask;
static G_POINT		bbox[8];
static int		last;

int
g_start_bird (x_wid, y_wid, z_wid, z_bottom, z_top, distance, theta, phi,
	      x_left_std, y_top_std, x_wid_std, y_wid_std,
	      sth, cth, sphi, cphi, last_axis, z_level, z_scale, proc)
    G_P_DIM	x_wid, y_wid, z_wid;
    G_P_REAL	z_bottom, z_top;
    G_P_DIM	distance;
    G_P_REAL	theta, phi;
    G_P_POS	x_left_std, y_top_std;
    G_P_DIM	x_wid_std, y_wid_std;
    G_P_REAL	*sth, *cth, *sphi, *cphi;
    int		*last_axis;
    G_P_POS	*z_level;
    G_P_REAL	*z_scale;
    char *	proc;
{
    G_POINT	bmax, bmin;
    G_P_REAL	ratio;
    G_P_POS	min_y;
    int		i;

    if (x_wid < 0.0 || y_wid < 0.0 || z_wid < 0.0)
    {
	g_warning (proc, "empty box");
	return G_YES;
    }

    if (x_wid_std < 0.0 || y_wid_std < 0.0 || distance < 0.0)
    {
	g_warning (proc, "invalid parameter specified");
	return G_YES;
    }

    saved_scale = * CURRENT (scale);
    saved_line = * CURRENT (line);
    saved_area = * CURRENT (area);
    saved_point = g_context.point;
    saved_bw_color_mask = g_context.bw_color_mask;

    theta *= PI / 180.0;
    phi *= PI / 180.0;
    * sth = sin (theta);
    * cth = cos (theta);
    * sphi = sin (phi);
    * cphi = cos (phi);

    /* the bounding box */
    SET2D (bmin, 0.0, 0.0);
    SET2D (bmax, 0.0, 0.0);
    min_y = 0.0;

    for (i = 0; i < 8; i ++)
    {
	G_POS	x = ((i & 1) ? x_wid : - x_wid) / 2.0;
	G_POS	y = ((i & 2) ? y_wid : - y_wid) / 2.0;
	G_POS	z = ((i & 4) ? z_wid : - z_wid) / 2.0;

	ROT (* sth, * cth, * sphi, * cphi, x, y, z);

	if (i == 0 || (i < 4 && y < min_y))
	{
	    min_y = y;
	    last = i;
	}

	PROJ (distance, x, y, z, bbox[i].x, bbox[i].y);
	SET2D (bmin, MIN (bbox[i].x, bmin.x), MIN (bbox[i].y, bmin.y));
	SET2D (bmax, MAX (bbox[i].x, bmax.x), MAX (bbox[i].y, bmax.y));
    }

    * last_axis = last;

    if (distance > 0.0 && min_y <= - distance)
    {
	g_warning (proc, "eye position too close");
	return G_YES;
    }

    if (bmax.x == bmin.x || bmax.y == bmin.y)
    {
	g_warning (proc, "projected area too small");
	return G_YES;
    }

    /* calculating the scale */
    ratio =  (bmax.y - bmin.y) / (bmax.x - bmin.x);

    if (ratio <= y_wid_std / x_wid_std)
    {
	y_top_std += (y_wid_std - x_wid_std * ratio) / 2.0;
	y_wid_std = x_wid_std * ratio;
    }
    else
    {
	x_left_std += (x_wid_std - y_wid_std / ratio) / 2.0;
	x_wid_std = y_wid_std / ratio;
    }

    g_set_scale (bmin.x, bmax.x, bmin.y, bmax.y,
		 x_left_std, y_top_std, x_wid_std, y_wid_std);

    * z_level = (z_top + z_bottom) / 2.0;
    * z_scale = z_wid / (z_top - z_bottom);

    /* drawing the bounding box */
    AXIS_LINE ();
    i = (phi >= 0.0) ? 0 : 4;
    Cg_move (bbox[i].x, bbox[i].y);
    Cg_plot (bbox[i + 1].x, bbox[i + 1].y);
    Cg_plot (bbox[i + 3].x, bbox[i + 3].y);
    Cg_plot (bbox[i + 2].x, bbox[i + 2].y);
    Cg_plot (bbox[i].x, bbox[i].y);

    for (i = 0; i < 4; i ++)
    {
	if (i != last)
	{
	    Cg_move (bbox[i].x, bbox[i].y);
	    Cg_plot (bbox[i + 4].x, bbox[i + 4].y);
	}
    }

    return G_NO;
}

void
g_end_bird ()
{
    /* last axis */
    AXIS_LINE ();
    Cg_move (bbox[last].x, bbox[last].y);
    Cg_plot (bbox[last + 4].x, bbox[last + 4].y);

    * CURRENT (scale) = saved_scale;
    * CURRENT (line) = saved_line;
    * CURRENT (area) = saved_area;
    g_context.point = saved_point ;
    g_set_bw_color (saved_bw_color_mask);
    CALL_PROC (validate_line, NULL);
    CALL_PROC (validate_area, NULL);
}

