
static char rcsid[] = "$Id: g_dev_x.c,v 1.5 1993/05/16 04:54:20 nakano Exp nakano $";

#include "g_private.h"
#include "g_dev_x.h"

static XFontStruct *
g_x_load_font (display, name)
    Display *	display;
    char * name;
{
    XFontStruct * font = XLoadQueryFont (display, name);

    if (font)
	return font;

    return XLoadQueryFont (display, "fixed");
}

static int
g_x_alloc_color (display, colormap, name)
    Display *	display;
    Colormap 	colormap;
    char *	name;
{
    XColor	screen_color;
    XColor	exact_color;

    if (! XAllocNamedColor (display, colormap, name,
			    & screen_color, & exact_color))
    {
	char	message[200];

	sprintf (message, "%s color not allowed for X, treated as black", name);
	g_warning ("g_device", message);

	return BlackPixelOfScreen (DefaultScreenOfDisplay (display));
    }

    return screen_color.pixel;
}

static G_BOOL
g_x_initialize (dev)
    DEVICE *	dev;
{
    X_PRIV *			priv = (X_PRIV *) dev -> dev_private;
    char *			disp_env;
    Display *			display;
    Screen *			screen;
    unsigned long		gc_mask;
    XGCValues   		gc_values;
    unsigned int		depth;
    Colormap 			colormap;
    G_POS			x_dpm, y_dpm;
    int				i;
    extern char *		getenv ();

    disp_env = getenv ("DISPLAY");

    if (disp_env == (char *) NULL)
    {
	g_warning ("g_device", "no DISPLAY environment for X");
	return G_YES;
    }

    priv -> display = display = XOpenDisplay (disp_env);

    if (display == (Display *) NULL)
    {
	char	message[200];

	sprintf (message, "can not open DISPLAY \"%s\" for X", disp_env);
	g_warning ("g_device", message);
	return G_YES;
    }

    screen = DefaultScreenOfDisplay (display);
    depth = DefaultDepthOfScreen (screen);
    colormap = DefaultColormapOfScreen (screen);

    x_dpm = (G_POS) WidthOfScreen (screen) / (G_POS) WidthMMOfScreen (screen); 
    y_dpm = (G_POS) HeightOfScreen (screen) / (G_POS) HeightMMOfScreen (screen);
    dev -> x_scale = dev -> y_scale = (x_dpm + y_dpm) / 2;

    if (priv -> window == (Window) NULL)
    {
	Window			window;
	int			width, height;
	XSizeHints		hints;
	unsigned long		attrib_mask;
	XSetWindowAttributes	attributes;
	XEvent			event;

	width = X_TRANSLATE_W (dev, g_context.win_width);
	height = X_TRANSLATE_H (dev, g_context.win_height);

	attributes.border_pixel = BlackPixelOfScreen (screen);
	attributes.background_pixel = WhitePixelOfScreen (screen);
	attributes.backing_store = Always;
	attrib_mask = CWBorderPixel | CWBackPixel | CWBackingStore;

	priv -> window = window
	    = XCreateWindow (display, RootWindowOfScreen (screen),
			     0, 0, width, height, 1, depth, InputOutput,
			     DefaultVisualOfScreen(screen),
			     attrib_mask, & attributes);

	hints.width = width;
	hints.height = height;
	hints.flags = USSize;
	XSetStandardProperties (display, window, "GLSC", (char *) NULL,
				(Pixmap) 0, NULL, 0, & hints);
	XSelectInput (display, window, StructureNotifyMask);
	XMapWindow (display, window);

	do {
	    XNextEvent (display, & event);
	} while (event.type != MapNotify);

	XSelectInput (display, window, NoEventMask);
	XSync (display, True);
    }
			
    gc_values.fill_rule = WindingRule;
    gc_mask = GCFillRule;
    priv -> line_gc = XCreateGC (display, RootWindowOfScreen (screen),
				 gc_mask, & gc_values);
    priv -> area_gc = XCreateGC (display, RootWindowOfScreen (screen),
				 gc_mask, & gc_values);
    priv -> text_gc = XCreateGC (display, RootWindowOfScreen (screen),
				 gc_mask, & gc_values);

    priv -> fonts[G_FONT_TIMES_8] = g_x_load_font (display, X_FONT_TIMES_8);
    priv -> fonts[G_FONT_TIMES_12] = g_x_load_font (display, X_FONT_TIMES_12);
    priv -> fonts[G_FONT_TIMES_18] = g_x_load_font (display, X_FONT_TIMES_18);
    priv -> fonts[G_FONT_TIMES_24] = g_x_load_font (display, X_FONT_TIMES_24);
    priv -> fonts[G_FONT_SYMBOL_8] = g_x_load_font (display, X_FONT_SYMBOL_8);
    priv -> fonts[G_FONT_SYMBOL_12] = g_x_load_font (display, X_FONT_SYMBOL_12);
    priv -> fonts[G_FONT_SYMBOL_18] = g_x_load_font (display, X_FONT_SYMBOL_18);
    priv -> fonts[G_FONT_SYMBOL_24] = g_x_load_font (display, X_FONT_SYMBOL_24);

    if (depth > 1)
    {
	priv -> palette[G_BLACK]
	    = g_x_alloc_color (display, colormap, "black" );
	priv -> palette[G_RED]
	    = g_x_alloc_color (display, colormap, "red" );
	priv -> palette[G_GREEN]
	    = g_x_alloc_color (display, colormap, "green" );
	priv -> palette[G_BLUE]
	    = g_x_alloc_color (display, colormap, "blue" );
	priv -> palette[G_MAGENTA]
	    = g_x_alloc_color (display, colormap, "magenta" );
	priv -> palette[G_YELLOW]
	    = g_x_alloc_color (display, colormap, "yellow" );
	priv -> palette[G_CYAN]
	    = g_x_alloc_color (display, colormap, "cyan" );
	priv -> palette[G_WHITE]
	    = g_x_alloc_color (display, colormap, "white" );
    }
    else
    {
	for (i = 0; i <= G_COLOR_MAX; i++)
	{
	    priv -> palette[i]
		= g_x_alloc_color (display, colormap,
				   (i == G_WHITE) ? "white" : "black" );
	}
    }

    dev -> initialize = (G_BOOL (*) G_ARGS ((DEVICE *))) NULL;

    return G_NO;
}

static void
g_x_terminate (dev)
    DEVICE *	dev;
{
    X_PRIV *		priv = (X_PRIV *) dev -> dev_private;

    if (priv -> display)
    {
	XCloseDisplay (priv -> display); 
	priv -> display = (Display *) NULL;
    }

    dev -> initialize = g_x_initialize;
}

static void
g_x_validate_bw_color (dev)
    DEVICE *	dev;
{
    X_PRIV *		priv = (X_PRIV *) dev -> dev_private;
    Screen *		screen = DefaultScreenOfDisplay (priv -> display);

    if (DefaultDepthOfScreen (screen) == 1)
    {
	Colormap 	colormap = DefaultColormapOfScreen (screen);
	unsigned long	mask = g_context.bw_color_mask;
	int		i;

	for (i = 0; i <= G_COLOR_MAX; i ++)
	{
	    priv -> palette[i]
		= g_x_alloc_color (priv -> display, colormap,
				   (mask & 0x1) ? "black" : "white");
	    mask = (mask >> 1);
	}
    }
}

static void
g_x_validate_line (dev)
    DEVICE *	dev;
{
    X_PRIV *		priv = (X_PRIV *) dev -> dev_private;
    LINE_PROP *		prop = CURRENT (line);
    unsigned long	gc_mask;
    XGCValues   	gc_values;
    char *		dashes = (char *) NULL;

    gc_mask = (GCForeground | GCLineStyle | GCLineWidth);

    switch (prop -> type)
    {
	case G_LINE_SOLID:
	    gc_values.line_style = LineSolid;
	    break;
	case G_LINE_DOTS:
	    gc_values.line_style = LineOnOffDash;
	    gc_values.dashes = 2;
	    gc_mask |= GCDashList;
	    break;
	case G_LINE_DASHED:
	    gc_values.line_style = LineOnOffDash;
	    gc_values.dashes = 4;
	    gc_mask |= GCDashList;
	    break;
	case G_LINE_LONG_DASHED:
	    gc_values.line_style = LineOnOffDash;
	    gc_values.dashes = 8;
	    gc_mask |= GCDashList;
	    break;
	case G_LINE_THIN_DOTS:
	    gc_values.line_style = LineOnOffDash;
	    dashes = "\004\014";
	    break;
	case G_LINE_DOT_DASHED:
	    gc_values.line_style = LineOnOffDash;
	    dashes = "\010\003\002\003";
	    break;
	case G_LINE_D_DOT_DASHED:
	    gc_values.line_style = LineOnOffDash;
	    dashes = "\010\002\001\002\001\002";
	    break;
	case G_LINE_RESERVED:
	    gc_values.line_style = LineSolid;
	    break;
    }

    gc_values.line_width = (prop -> width <= 1) ? 0 : prop -> width;
    gc_values.foreground = priv -> palette[prop -> color];
    XChangeGC (priv -> display, priv -> line_gc, gc_mask, & gc_values);

    if (dashes)
    {
	XSetDashes (priv -> display, priv -> line_gc,
		    0, dashes, strlen (dashes));
    }
}

static void
g_x_validate_area (dev)
    DEVICE *	dev;
{
    X_PRIV *	priv = (X_PRIV *) dev -> dev_private;

    XSetForeground
	(priv -> display, priv -> area_gc,
	 (unsigned long) priv -> palette[CURRENT (area) -> color]);
}

static void
g_x_validate_text (dev)
    DEVICE *	dev;
{
    X_PRIV *		priv = (X_PRIV *) dev -> dev_private;
    TEXT_PROP *		prop = CURRENT (text);
    unsigned long	gc_mask;
    XGCValues   	gc_values;

    gc_mask = (GCForeground | GCFont);
    gc_values.foreground = priv -> palette[prop -> color];
    gc_values.font = priv -> fonts[prop -> font] -> fid;
    XChangeGC (priv -> display, priv -> text_gc, gc_mask, & gc_values);
}

static void
g_x_clear (dev)
    DEVICE *	dev;
{
    X_PRIV *		priv = (X_PRIV *) dev -> dev_private;

    XClearWindow (priv -> display, priv -> window);
}

static void
g_x_line (dev, packet)
    DEVICE *	dev;
    OPAQUE	packet;
{
    X_PRIV *		priv = (X_PRIV *) dev -> dev_private;
    G_LINE_DEF *	p = (G_LINE_DEF *) packet;

    XDrawLine (priv -> display, priv -> window, priv -> line_gc,
	       X_TRANSLATE_X (dev, p -> start.x),
	       X_TRANSLATE_Y (dev, p -> start.y),
	       X_TRANSLATE_X (dev, p -> end.x),
	       X_TRANSLATE_Y (dev, p -> end.y));
    XFlush (priv -> display);
}

static void
g_x_polyline (dev, packet)
    DEVICE *	dev;
    OPAQUE	packet;
{
    X_PRIV *		priv = (X_PRIV *) dev -> dev_private;
    G_POLYLINE_DEF *	p = (G_POLYLINE_DEF *) packet;
    XPoint		points[G_MAX_N_POINTS];
    int			i;

    if (p -> n_points == 1)
    {
	XDrawPoint (priv -> display, priv -> window, priv -> line_gc,
		    X_TRANSLATE_X (dev, p -> points[0].x),
		    X_TRANSLATE_Y (dev, p -> points[0].y));
    }
    else
    {
	for (i = 0; i < p -> n_points; i ++)
	{
	    points[i].x = X_TRANSLATE_X (dev, p -> points[i].x);
	    points[i].y = X_TRANSLATE_Y (dev, p -> points[i].y);
	}

	XDrawLines (priv -> display, priv -> window, priv -> line_gc,
		    points, p -> n_points, CoordModeOrigin);
    }

    XFlush (priv -> display);
}

static void
g_x_polygon (dev, packet)
    DEVICE *	dev;
    OPAQUE	packet;
{
    X_PRIV *		priv = (X_PRIV *) dev -> dev_private;
    G_POLYGON_DEF *	p = (G_POLYGON_DEF *) packet;
    XPoint		points[G_MAX_N_POINTS + 1];
    int			i;

    for (i = 0; i < p -> n_points; i ++)
    {
	points[i].x = X_TRANSLATE_X (dev, p -> points[i].x);
	points[i].y = X_TRANSLATE_Y (dev, p -> points[i].y);
    }

    points[p -> n_points] = points[0];

    if (p -> fill)
    {
	XFillPolygon (priv -> display, priv -> window, priv -> area_gc,
		      points, p -> n_points + 1, Complex, CoordModeOrigin);
    }

    if (p -> edge)
    {
	XDrawLines (priv -> display, priv -> window, priv -> line_gc,
		    points, p -> n_points + 1, CoordModeOrigin);
    }

    XFlush (priv -> display);
}

static void
g_x_circle (dev, packet)
    DEVICE *	dev;
    OPAQUE	packet;
{
    X_PRIV *		priv = (X_PRIV *) dev -> dev_private;
    G_CIRCLE_DEF *	p = (G_CIRCLE_DEF *) packet;
    int			x, y;
    unsigned int	w, h;

    w = X_TRANSLATE_W (dev, p -> radius_x);
    x = X_TRANSLATE_X (dev, p -> x) - w;
    y = X_TRANSLATE_Y (dev, p -> y) - w;
    w = h = w * 2;

    if (p -> fill)
    {
	XFillArc (priv -> display, priv -> window, priv -> area_gc,
		  x, y, w, h, 0, 360 * 64);
    }

    if (p -> edge)
    {
	XDrawArc (priv -> display, priv -> window, priv -> line_gc,
		  x, y, w, h, 0, 360 * 64);
    }

    XFlush (priv -> display);
}

static void
g_x_text (dev, packet)
    DEVICE *	dev;
    OPAQUE	packet;
{
    X_PRIV *		priv = (X_PRIV *) dev -> dev_private;
    G_TEXT_DEF *	p = (G_TEXT_DEF *) packet;

    XDrawString (priv -> display, priv -> window, priv -> text_gc,
		 X_TRANSLATE_X (dev, p -> x), X_TRANSLATE_Y (dev, p -> y),
		 p -> string, strlen (p -> string));
    XFlush (priv -> display);
}

static void
g_x_sleep (dev, packet)
    DEVICE *	dev;
    OPAQUE	packet;
{
    X_PRIV *	priv = (X_PRIV *) dev -> dev_private;
    G_REAL	time = ((G_SLEEP_DEF *) packet) -> time;

    if (time > 0.0)
    {
	g_msleep ((int) (time * 1000));
    }
    else if (time < 0.0)
    {
	XEvent	event;

	XSelectInput (priv -> display, priv -> window, ButtonPressMask);

	do {
	    XNextEvent (priv -> display, & event);
	} while (event.type != ButtonPress);

	XSelectInput (priv -> display, priv -> window, NoEventMask);
	XSync (priv -> display, True);
    }
}

static X_PRIV		g_x_dev_private;

void
g_x_sel_window (display, window)
    Display *	display;
    Window	window;
{
    unsigned long		attrib_mask;
    XSetWindowAttributes	attributes;

    g_x_dev_private.window = window;

    attributes.backing_store = Always;
    attrib_mask = CWBackingStore;
    XChangeWindowAttributes (display, window, attrib_mask, & attributes);
}

DEVICE	g_x_device =
{
    /* defined			*/	G_YES,
    /* x_scale			*/	1.0,
    /* y_scale			*/	1.0,
    /* initialize		*/	g_x_initialize,
    /* terminate		*/	g_x_terminate,
    /* validate_bw_color	*/	g_x_validate_bw_color,
    /* validate_line		*/	g_x_validate_line,
    /* validate_area		*/	g_x_validate_area,
    /* validate_text		*/	g_x_validate_text,
    /* clear			*/	g_x_clear,
    /* line			*/	g_x_line,
    /* polyline			*/	g_x_polyline,
    /* polygon			*/	g_x_polygon,
    /* circle			*/	g_x_circle,
    /* text			*/	g_x_text,
    /* sleep			*/	g_x_sleep,
    /* dev_private		*/	(OPAQUE) & g_x_dev_private,
};
