
static char rcsid[] = "$Id: g_f_hidden.c,v 1.4 1995/05/23 05:26:11 nakano Exp nakano $";

#include "g_private.h"
#include "g_bird.h"

#define SPAN(n, i, intv) \
	((((n) - (i)) % (intv)) ? ((n) - (i)) % (intv) : (intv))

static void
side (side_mode, data, n_x, n_y, unit_x, unit_y,
      x_pos, y_pos, dx, dy, z_pos, y_to_x, y_to_y)
    G_SIDE_MODE	side_mode;
    G_REAL *	data;
    int		n_x, n_y, unit_x, unit_y;
    G_P_POS	x_pos, y_pos, dx, dy, z_pos;
    G_P_REAL	y_to_x, y_to_y;
{
    G_POS	gx[G_MAX_N_POINTS + 2], gy[G_MAX_N_POINTS + 2];
    int		n, unit, i, j;

    if (n_x)
    {
	if ((side_mode & G_SIDE_Y) == 0)
	    return;

	n = n_x;
	unit = unit_x;
	dy = 0.0;
    }
    else if (n_y)
    {
	if ((side_mode & G_SIDE_X) == 0)
	    return;

	n = n_y;
	unit = unit_y;
	dx = 0.0;
    }
    else
    {
	return;
    }

    gx[0] = FAKE_X (x_pos, y_pos, z_pos, y_to_x, y_to_y);
    gy[0] = FAKE_Y (x_pos, y_pos, z_pos, y_to_x, y_to_y);

    j = 0;

    for (i = 1; i <= n; i ++)
    {

	gx[i] = FAKE_X (x_pos, y_pos, data[j], y_to_x, y_to_y);
	gy[i] = FAKE_Y (x_pos, y_pos, data[j], y_to_x, y_to_y);
	j += unit;
	x_pos += dx;
	y_pos += dy;
    }

    x_pos -= dx, y_pos -= dy;
    gx[i] = FAKE_X (x_pos, y_pos, z_pos, y_to_x, y_to_y);
    gy[i] = FAKE_Y (x_pos, y_pos, z_pos, y_to_x, y_to_y);

    SIDE_AREA ();
    Cg_polygon (gx, gy, n + 2, G_YES, G_YES);
}

static void
patch (data, unit_x, unit_y, x_pos, y_pos, dx, dy, n_x, n_y, y_to_x, y_to_y)
    G_REAL *	data;
    int		unit_x, unit_y;
    G_P_POS	x_pos, y_pos, dx, dy;
    int		n_x, n_y;
    G_P_REAL	y_to_x, y_to_y;
{
    G_POS	gx[G_MAX_N_POINTS], gy[G_MAX_N_POINTS];
    int		i, j;

    j = 0;

    for (i = 0; i < n_x; i ++)
    {
	gx[j] = FAKE_X (x_pos, y_pos, * data, y_to_x, y_to_y);
	gy[j] = FAKE_Y (x_pos, y_pos, * data, y_to_x, y_to_y);
	x_pos += dx, data += unit_x, j ++;
    }

    for (i = 0; i < n_y; i ++)
    {
	gx[j] = FAKE_X (x_pos, y_pos, * data, y_to_x, y_to_y);
	gy[j] = FAKE_Y (x_pos, y_pos, * data, y_to_x, y_to_y);
	y_pos += dy, data += unit_y, j ++;
    }

    for (i = 0; i < n_x; i ++)
    {
	gx[j] = FAKE_X (x_pos, y_pos, * data, y_to_x, y_to_y);
	gy[j] = FAKE_Y (x_pos, y_pos, * data, y_to_x, y_to_y);
	x_pos -= dx, data -= unit_x, j ++;
    }

    for (i = 0; i < n_y; i ++)
    {
	gx[j] = FAKE_X (x_pos, y_pos, * data, y_to_x, y_to_y);
	gy[j] = FAKE_Y (x_pos, y_pos, * data, y_to_x, y_to_y);
	y_pos -= dy, data -= unit_y, j ++;
    }

    EDGE_LINE ();
    ROOF_AREA ();
    Cg_polygon (gx, gy, j, G_YES, G_YES);
}

static void
do_hidden (z_bottom, z_top, x_ratio, y_ratio,
	   x_left_std, y_top_std, x_wid_std, y_wid_std,
	   data, n_x, n_y, unit_x, unit_y,
	   direction, side_mode, intv_x, intv_y)
    G_P_REAL	z_bottom, z_top;
    G_P_REAL	x_ratio, y_ratio;
    G_P_POS	x_left_std, y_top_std;
    G_P_DIM	x_wid_std, y_wid_std;
    G_REAL *	data;
    int		n_x, n_y, unit_x, unit_y, direction;
    G_SIDE_MODE	side_mode;
    int		intv_x, intv_y;
{
    G_P_POS	x, y, dx, dy;
    G_P_REAL	y_to_x, y_to_y;
    G_REAL *	y_data;
    G_REAL *	x_data;
    int		span_x, span_y;
    int		i, j;

    g_check_init ();

    if ((intv_x + intv_y) * 2 > G_MAX_N_POINTS)
    {
	char	message[200];
	int	old = intv_x;

	intv_x = G_MAX_N_POINTS / 2 / intv_y;
	intv_x = n_x / ((n_x + intv_x - 1) / intv_x);
	sprintf (message, "too short x-interval %d adjusted to %d",
		 old, intv_x);
	g_warning ("g_hidden", message);
    }

    if (g_start_fake_bird (z_bottom, z_top, x_ratio, y_ratio,
			   x_left_std, y_top_std, x_wid_std, y_wid_std,
			   & y_to_x, & y_to_y, "g_fake_hidden"))
    {
	return;
    }

    SET_BW_COLOR ();

    if (direction < 0)
    {
	data += unit_y * (n_y - 1); 
	unit_y = - unit_y;
    }

    dx = 1.0 / (n_x - 1), dy = 1.0 / (n_y - 1);

    side (side_mode, data, 0, n_y, 0, unit_y,
	  (G_POS) 0.0, (G_POS) 0.0, (G_POS) 0.0, dy, z_bottom, y_to_x, y_to_y);
    side (side_mode, data + unit_y * (n_y - 1), n_x, 0, unit_x, 0,
	  (G_POS) 0.0, (G_POS) 1.0, dx, (G_POS) 0.0, z_bottom, y_to_x, y_to_y);

    y = 1.0;
    y_data = data + unit_y * (n_y - 1);

    for (i = 1; i < n_y; i += span_y)
    {
	x = 0.0;
	x_data = y_data;
	span_y = SPAN (n_y, i, intv_y);

	for (j = 1; j < n_x; j += span_x)
	{
	    span_x = SPAN (n_x, j, intv_x);
	    patch (x_data, unit_x, - unit_y, x, y, dx, - dy,
		   span_x, span_y, y_to_x, y_to_y);
	    x_data += unit_x * span_x;
	    x += dx * span_x;
	}

	y_data -= unit_y * span_y;
	y -= dy * span_y;
    }

    side (side_mode, data, n_x, 0, unit_x, 0,
	  (G_POS) 0.0, (G_POS) 0.0, dx, (G_POS) 0.0, z_bottom, y_to_x, y_to_y);
    side (side_mode, data + unit_x * (n_x - 1), 0, n_y, 0, unit_y,
	  (G_POS) 1.0, (G_POS) 0.0, (G_POS) 0.0, dy, z_bottom, y_to_x, y_to_y);

    g_end_fake_bird ();
}

void
Cg_fake_hidden (z_bottom, z_top, x_ratio, y_ratio,
	       x_left_std, y_top_std, x_wid_std, y_wid_std,
	       data, n_x, n_y, direction, side_mode, intv_x, intv_y)
    G_P_REAL	z_bottom, z_top;
    G_P_REAL	x_ratio, y_ratio;
    G_P_POS	x_left_std, y_top_std, x_wid_std, y_wid_std;
    G_REAL *	data;
    int		n_x, n_y, direction, side_mode, intv_x, intv_y;
{
    do_hidden (z_bottom, z_top, x_ratio, y_ratio,
	       x_left_std, y_top_std, x_wid_std, y_wid_std,
	       data, n_x, n_y, n_y, 1,
	       direction, side_mode, intv_x, intv_y);
}

void
F77g_fake_hidden (z_bottom, z_top, x_ratio, y_ratio,
		x_left_std, y_top_std, x_wid_std, y_wid_std,
		data, n_x, n_y, direction, side_mode, intv_x, intv_y)
    G_REAL	*z_bottom, *z_top;
    G_REAL	*x_ratio, *y_ratio;
    G_POS	*x_left_std, *y_top_std, *x_wid_std, *y_wid_std;
    G_REAL	*data;
    int		*n_x, *n_y, *direction;
    G_SIDE_MODE	*side_mode;
    int		*intv_x, *intv_y;
{
    do_hidden (*z_bottom, *z_top, *x_ratio, *y_ratio,
	       *x_left_std, *y_top_std, *x_wid_std, *y_wid_std,
	       data, *n_x, *n_y, 1, *n_x,
	       *direction, *side_mode, *intv_x, *intv_y);
}

void
g_fake_hidden__ (z_bottom, z_top, x_ratio, y_ratio,
		x_left_std, y_top_std, x_wid_std, y_wid_std,
		data, n_x, n_y, direction, side_mode, intv_x, intv_y)
    G_REAL	*z_bottom, *z_top;
    G_REAL	*x_ratio, *y_ratio;
    G_POS	*x_left_std, *y_top_std, *x_wid_std, *y_wid_std;
    G_REAL	*data;
    int		*n_x, *n_y, *direction;
    G_SIDE_MODE	*side_mode;
    int		*intv_x, *intv_y;
{
    do_hidden (*z_bottom, *z_top, *x_ratio, *y_ratio,
	       *x_left_std, *y_top_std, *x_wid_std, *y_wid_std,
	       data, *n_x, *n_y, 1, *n_x,
	       *direction, *side_mode, *intv_x, *intv_y);
}

