
static char rcsid[] = "$Id: g_sleep.c,v 1.5 1995/05/23 05:26:27 nakano Exp nakano $";

#include "g_private.h"

#if G_OS_USLEEP
#define USE_USLEEP
#else
#undef USE_USLEEP
#if G_OS_SETITIMER && G_OS_SIGHOLD
#define USE_SETITIMER
#else
#undef USE_SETITIMER
#if G_OS_TERMIO && G_OS_PTY
#define USE_TERMIO
#else
#undef USE_TERMIO
#define USE_SLEEP
#endif
#endif
#endif

#ifdef USE_SETITIMER
#include <sys/time.h>
#include <signal.h>
static void
handle_alarm ()
{
}
#endif

#ifdef USE_TERMIO
#include <termio.h>
#include <fcntl.h>
#endif

void
g_msleep (msec)
    int	msec;
{
#ifdef USE_USLEEP
    if (msec > 0)
    {
	if (msec >= 1000)
	    sleep ((unsigned) (msec / 1000));

	usleep ((unsigned) (msec % 1000 * 1000));
    }
#endif
#ifdef USE_SETITIMER
#define SUBTRACT_IT(a, b) \
    { \
	if (a.it_value.tv_usec < b.it_value.tv_usec) \
	{ \
	    a.it_value.tv_sec --; \
	    a.it_value.tv_usec += 1000000; \
	} \
	a.it_value.tv_usec -= b.it_value.tv_usec; \
	a.it_value.tv_sec -= b.it_value.tv_sec; \
    }

    if (msec > 0)
    {
	void 			(* old_handler) ();
	struct itimerval	new, old, zero;

	new.it_value.tv_sec = msec / 1000;
	new.it_value.tv_usec = (msec % 1000) * 1000;
	new.it_interval.tv_sec = 0;
	new.it_interval.tv_usec = 0;
	zero.it_value.tv_sec = zero.it_value.tv_usec = 0;
	zero.it_interval.tv_sec = zero.it_interval.tv_usec = 0;
	setitimer (ITIMER_REAL, & zero, & old);

	while ((old.it_value.tv_sec != 0 || old.it_value.tv_usec != 0)
	       && (old.it_value.tv_sec < new.it_value.tv_sec
		   || (old.it_value.tv_sec == new.it_value.tv_sec
		       && old.it_value.tv_usec <= new.it_value.tv_usec)))
	{
	    SUBTRACT_IT (new, old);
	    setitimer (ITIMER_REAL, & old, (struct itimerval *) 0);
	    sigpause (SIGALRM);
	    setitimer (ITIMER_REAL, & zero, & old);
	}

	if (new.it_value.tv_sec != 0 || new.it_value.tv_usec != 0)
	{
	    old_handler = signal (SIGALRM, handle_alarm);
	    sighold (SIGALRM);
	    setitimer (ITIMER_REAL, & new, (struct itimerval *) 0);
	    sigpause (SIGALRM);
	    signal (SIGALRM, old_handler);
	}

	if (old.it_value.tv_sec != 0 || old.it_value.tv_usec != 0)
	{
	    SUBTRACT_IT (old, new);
	    setitimer (ITIMER_REAL, & old, (struct itimerval *) 0);
	}
    }
#endif
#ifdef USE_TERMIO
    static int			master = -1, slave = -1;
    static struct termio	term;
    char			c = ' ';	/* a dummy character */
    char			buf[10];
    int				sec = msec / 1000;
    int				dsec = (msec % 1000 + 50) / 100;

    if (msec <= 0)
	return;

    if (slave == -1 || master == -1)
    {
	int	i, done = G_NO;
	char	c, name[100];

	for (c = 'p'; ! done && c <= 's'; c ++) {
	    for (i = 0; ! done && i < 16; i ++) {
		sprintf (name, "/dev/pty%c%x", c, i);
		if ((master = open (name, O_RDWR)) < 0)
		    continue;

		sprintf (name, "/dev/tty%c%x", c, i);
		if ((slave = open (name, O_RDWR)) < 0) {
		    close (master);
		    master = -1;
		    continue;
		}
		done = G_YES;
	    }
	}

	if (! done) {
	    sleep ((unsigned) ((msec + 500) / 1000));
	    return;
	}

	ioctl (slave, TCGETA, & term);
	term.c_iflag &= ~(INLCR | ICRNL | IXON | BRKINT);
	term.c_oflag &= ~(OPOST);
	term.c_lflag &= ~(ICANON | ISIG | ECHO);
	term.c_cc[4] = 10;
    }

    if (dsec > 0)
    {
	term.c_cc[5] = dsec;
	ioctl (slave, TCSETA, & term);
	write (master, & c, 1);
	read (slave, buf, sizeof (buf));
    }

    if (sec > 0)
	sleep ((unsigned) sec);
#endif
#ifdef USE_SLEEP
    if (msec > 0)
	sleep ((unsigned) ((msec + 500) / 1000));
#endif
}

void
Cg_sleep (time)
    G_P_REAL	time;
{
    G_SLEEP_DEF	packet;

    g_check_init ();
    packet.time = time;
    CALL_PROC (sleep, & packet);
}

void
F77g_sleep (time)
    G_REAL *	time;
{
    Cg_sleep (* time);
}

void
g_sleep__ (time)
    G_REAL *	time;
{
    Cg_sleep (* time);
}

